// use phylodm::PDM;

fn main() {
    // let mut tree = PDM::default();
    // tree.load_from_newick_path("/Users/aaron/tmp/500_taxa.tree");
    // // tree.load_from_newick_path("/Users/aaron/git/PhyloDM/scripts/generate_test_data/test.tree");
    //
    // // tree.load_from_newick_path(&"/Users/aaron/git/PhyloDM/src/newick.txt".to_string());
    // let _arr = tree.matrix(false);
    //
    // return;
}
